(***************************************************************************
 *
 * Index Set Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   27 October 2001
 *
 ***************************************************************************)

structure IndexSet : SET =
  struct
    (* indexsets are implemented as sorted sets of strings *)
    structure StringElem =
      struct
	type t = String.string
	val eq = op=
	val op< = String.<
	val toString = String.toString
    end

    structure SetImpl : SET = Set (structure Elem = StringElem)

    type t = StringElem.t
    type 'a set = 'a SetImpl.set

    (* the results of the following functions don't need to be sorted *)
    val empty = SetImpl.empty
    val isEmpty = SetImpl.isEmpty
    val elem = SetImpl.elem
    val subset = SetImpl.subset
    val eq = SetImpl.eq
    val powerSet = SetImpl.powerSet
    val crossProduct = SetImpl.crossProduct
    val setMap = SetImpl.setMap

    (* sort s
     *   uses Quicksort to sort the string list s
     *)
    fun sort nil = nil
      | sort (h::t) =
	    sort (List.filter (fn x => StringElem.< (x, h)) t)
	      @ [h] @ sort (List.filter (fn x => StringElem.< (h, x)) t)

    (* union (s1, s2)
     *   returns the union of indexsets s1 and s2
     *   Ex: union (["i", "j"], ["j", "k"]) returns ["i", "j", "k"]
     *)
    fun union (s1, s2) = sort (SetImpl.union (s1, s2))

    (* intersect (s1, s2)
     *   returns the intersection of indexsets s1 and s2
     *   Ex: intersect (["i", "j"], ["j", "k"]) returns ["j"]
     *)
    fun intersect (s1, s2) = sort (SetImpl.intersect (s1, s2))

    (* Union s
     *   returns the union of indexsets in list s
     *   Ex: Union [["i", "j"], ["j", "k"]] returns ["i", "j", "k"]
     *)
    fun Union s = sort (SetImpl.Union s)

    (* Intersect s
     *   returns the intersection of indexsets in list s
     *   Ex: Intersect [["i", "j"], ["j", "k"]] returns ["j"]
     *)
    fun Intersect s = sort (SetImpl.Intersect s)

    (* difference (s1, s2)
     *   returns the set of all the indices from s1 that are not in s2
     *   Ex: difference (["i", "j", "k"], ["j", "k"]) returns ["i"]
     *)
    fun difference (s1, s2) = sort (SetImpl.difference (s1, s2))

    (* toString s
     *   constructs a string out of an indexset by concatenating all indices
     *   Ex: toString ["i", "j", "k"] returns "ijk"
     *)
    fun toString s = concat (sort s)
  end
