(***************************************************************************
 *
 * Index Set Sequence Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   4 September 2001
 *
 ***************************************************************************)

structure IndexSetSeq : INDEX_SET_SEQ =
  struct
    type t = IndexSet.t
    type iSet = t IndexSet.set
    type iSetSeq = iSet list

    structure IndexSetElem =
      struct
	type t = IndexSet.t IndexSet.set
	val eq = IndexSet.eq
	val toString = IndexSet.toString
      end

    structure ISetSet = Set (structure Elem = IndexSetElem)

    (* toString f
     *   constructs a string out of an indexset sequence
     *   Ex: [["i", "j"], ["k"]] is translated to "<ij, k>"
     *)
    local
      fun toString' nil = ""
	| toString' [s] = IndexSet.toString s
	| toString' (h::t) = IndexSet.toString h ^ ", " ^ toString' t
    in
      fun toString f = "<" ^ toString' f ^ ">"
    end

    structure ISetSeqSet =
      struct
	type t = iSetSeq
	val eq = op=
	val toString = toString
      end

    structure ISetSeqSet = Set (structure Elem = ISetSeqSet)

    type iSetSeqSet = ISetSeqSet.t ISetSeqSet.set

    (* any
     *   the undefined nesting, allows arbitrary fusion with a parent
     *)
    val any = [["*"]]

    (* set f
     *   returns the set of all the indices in the indexset sequence f
     *   Ex: set [["i", "j"], ["k"]] returns ["i", "j", "k"]
     *)
    fun set f = foldr IndexSet.union nil f

    (* prefix (f1, f2)
     *   returns true iff the sequence f1 is a prefix of the sequence f2
     *   Ex: prefix ([["i"], ["j"]], [["i"], ["j", "k"]]) returns true
     *)
    fun prefix (nil, _) = true
      | prefix (_::_, nil) = false
      | prefix ([h1], h2::_) = IndexSet.subset (h1, h2)
      | prefix (h1::t1, h2::t2) = h1 = h2 andalso prefix (t1, t2)

    (* prefixes f
     *   returns the list of all prefixes of fusion f
     *)
    fun prefixes nil = [nil]
      | prefixes (h::t) =
	  ISetSeqSet.union (map (fn s => [s])
				(ISetSet.difference (IndexSet.powerSet h,
						     [nil])),
			    map (fn s => h :: s)
				(prefixes t))
	    @ [[]]

    (* maxPrefix (f, s)
     *   returns the maximum prefix of indexset sequence f that contains
     *   only elements in s
     *   Ex: maxPrefix ([["i"], ["j"], ["k", "l"], ["m"]], ["i", "j", "k"])
     *       returns [["i"], ["j"], ["k"]]
     *)
    fun maxPrefix (nil, s) = nil
      | maxPrefix (h::t, s) =
	  if (IndexSet.subset (h, s)) then h :: maxPrefix (t, s)
	  else let val h' = List.filter (fn x => IndexSet.elem (x, s)) h
	       in
		 if h' = nil then nil else [h']
	       end

    (* extend (f, s)
     *   extends the indexset sequence f with all the indices in s that
     *   are not in (set f)
     *   Ex: extend ([["i", "j"], ["k"]], ["j", "l", "m"])
     *       returns [["i", "j"], ["k"], ["l", "m"]]
     *)
    fun extend (nil, nil) = nil
      | extend (nil, s) = [s]
      | extend (f, nil) = if f = any then nil else f
      | extend (f, s) =
	  if f = any then [s]
	  else
	    let val d = IndexSet.difference (s, set f)
	    in
	      if d = nil then f else f @ [d]
	    end

    (* exception Incompatible
     *   indicates that the arguments of merge are incompatible
     *)
    exception Incompatible

    local
      fun merge' (nil, nil) = nil
	| merge' (nil, _::_) = nil
	| merge' (_::_, nil) = nil
	| merge' (f as (h::t), f' as (h'::t')) =
	    let
	      val y = IndexSet.intersect(h, h')
	      val x = IndexSet.difference (h, y)
	      val x' = IndexSet.difference (h', y)
	    in
	      if y = nil then nil
	      else if x = nil andalso x' = nil then [y] @ merge' (t, t')
	      else if x = nil then [y] @ merge' (t, x'::t')
	      else if x' = nil then [y] @ merge' (x::t, t')
	      else [y] @ merge' (x::t, x'::t')
	    end
    in
      (* merge (f1, f2)
       *   returns the least constraining indexset sequence r that is more or
       *   equally constraining than both f1 and f2, returns nil if the
       *   arguments are not compatible
       *
       *   requires: set f1 = set f2
       *   ensures:  set r = set f1 andalso leq (r, f1) andalso leq (r, f2)
       *             andalso forall r'. (set r' = set r
       *			     andalso leq (r', f1) andalso leq (r', f2))
       *				  implies leq (r', r)
       *   Ex: merge ([["i", "j"], ["k"]], [["i"], ["j", "k"]])
       *       returns [["i"], ["j"], ["k"]]
       *)
      fun merge (nil, nil) = nil
	| merge (f, f') =
	  (* If either f or f' is any that should be a compiler error. *)
	  if f = any orelse f' = any orelse set f <> set f' then
	    raise Incompatible
	  else
	    let val m = merge' (f, f')
	    in
	      if set m = set f then m else raise Incompatible
	    end

      (* restrict (f1, f2) =
       *   restrict fusion f1 to the constraints from the parent's fusion f2
       *)
      fun restrict (f1, f2) =
	if f1 = any then f2
	else merge' (f1, extend (f2, set f1))

      (* leq (f1, f2)
       *   returns true iff f1 and f2 contain the same indices and
       *   f1 is equal to f2 or more constraining than f2,
       *   Ex: leq ([["i", "j"], ["k"]], [["i", "j", "k"]]) returns true
       *)
      fun leq (h, h') = set h = set h' andalso h = merge' (h, h')
    end

    (* suffix (h, f)
     *   returns the suffix of h under f by deleting all index sets from
     *   h that are in f;
     *   Ex: suffix ([["i", "j"], ["k"]], [["i", "j"]]) returns [["k"]]
     *)
    fun suffix (h, f) =
      ISetSet.difference (restrict (h, f), f)

  end
