(***************************************************************************
 *
 * Solution Tree Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   10 March 2002
 * Date:   18 March 2010
 *
 ***************************************************************************)

structure SolTree : SOL_TREE =
  struct

    type iSet = Solution.iSet
    type iSetSeq = Solution.iSetSeq
    type expTree = ExpTree.expTree
    type sol = Solution.sol

    (* datatype solTree = Leaf of sol
     *			| Extended of sol * iSetSeq * solTree
     *			| Reduced of sol * iSet * solTree
     *			| Unary of sol * solTree
     *			| Binary of sol * solTree * solTree
     * Ex:  Unary (Solution.S ([["i", "j"]], (1, 1, 0, 0), [], 0, f1),
     *	       Leaf (Solution.S ([["i", "j"]], (1, 1, 0, 0), [], 0, A)))
     *
     * Leaf s
     *   constructs a leaf node out of the solution s, which must contain an
     *   Array node, a Fun node, or an IntLit node
     * Extended (s, f, l)
     *   constructs an extended solution out of a given solution s and a
     *   nesting that was extended to the parent's mesting; extended
     *   solutions are constructed when the parent node has more indices
     *   than the child node
     * Reduced (s, i, l)
     *   constructs a reduced solution, a solution whose nesting was reduced
     *   from a given solution s to a smaller set of outer loops i, and whose
     *   cost was updated correspondingly; reduced solutions are constructed
     *   for hiding summation indices and at the top level
     * Unary (s, l)
     *   constructs a unary tree node out of the solution s, which must contain
     *   an Output node, a Cast node, or a Sum node, and the subtree l
     * Binary (s, l, r)
     *   constructs a binary tree node out of the solution s, which must
     *   contain a BinOp node, and the subtrees l and r
     *)
    datatype solTree = Leaf of sol
		     | Extended of sol * iSetSeq * solTree
		     | Reduced of sol * iSet * solTree
		     | Unary of sol * solTree
		     | Binary of sol * solTree * solTree

    local
      (* returns a string of n spaces *)
      fun indent n = if n = 0 then "" else " " ^ indent (n - 1)

      (* creates a string representation of a solution tree t indented
	 by n spaces *)
      fun toString' (Leaf s, n) =
	    indent n ^ Solution.toString s
	| toString' (Extended (s, f, l), n) =
	    indent n ^ "E" ^ IndexSetSeq.toString f ^ " "
	      ^ Solution.toString s ^ toString' (l, n + 2)
	| toString' (Reduced (s, i, l), n) =
	    indent n ^ "R<" ^ IndexSet.toString i ^ "> "
	      ^ Solution.toString s ^ toString' (l, n + 2)
	| toString' (Unary (s, l), n) =
	    indent n ^ Solution.toString s ^ toString' (l, n + 2)
	| toString' (Binary (s, l, r), n) =
	    indent n ^ Solution.toString s
	      ^ toString' (l, n + 2) ^ toString' (r, n + 2)
    in
      (* toString t
       *   creates a string representation of a solution tree t
       *)
      fun toString t = toString' (t, 0)
    end

    structure SolTreeElem =
      struct
	type t = solTree
	val eq = op=
	val toString = toString
      end

    structure SolTreeSet = Set (structure Elem = SolTreeElem)

    type solTreeSet = SolTreeSet.t SolTreeSet.set

    (* getSoln t
     *   returns the solution out of a solution tree node
     *)
    fun getSoln (Leaf s) = s
      | getSoln (Extended (s, _, _)) = s
      | getSoln (Reduced (s, _, _)) = s
      | getSoln (Unary (s, _)) = s
      | getSoln (Binary (s, _, _)) = s

    (* getExtSoln t
     *   returns the extended solution out of a solution tree node
     *)
    fun getExtSoln (Leaf s) = s
      | getExtSoln (Extended (Solution.S (f, c, frags, rc, t), f', _)) =
	Solution.S (f', c, frags, rc, t)
      | getExtSoln (Reduced (s, _, _)) = s
      | getExtSoln (Unary (s, _)) = s
      | getExtSoln (Binary (s, _, _)) = s

    local
      (* pruneSoln (s, S1)
       *   adds a solution tree s to solution tree set S1
       *)
      fun pruneSoln (s, S1) =
	let val s' = getExtSoln s
	in
	  if (List.exists (fn x => Solution.inferior (s', getExtSoln x)) S1)
	  then
	    S1
	  else
	    s::(List.filter (fn x => not (Solution.inferior (getExtSoln x,s')))
			    S1)
	end

      (* pruneSolns' (S1, S2)
       *   adds the solutions in S1 to S2
       *)
      fun pruneSolns' (nil, S2) = S2
	| pruneSolns' (h::t, S2) =
	    let val Solution.S (_, (c, _, _, _), _, _, _) = getSoln h
	    in
	      if Solution.exceedsMemoryLimit c then pruneSolns' (t, S2)
	      else pruneSolns' (t, pruneSoln (h, S2))
	    end
    in
      (* pruneSolns S1
       *   prune inferior solutions from solution set S1
       *)
      fun pruneSolns S1 = pruneSolns' (S1, [])
    end

    (* reduceSolns (S1, i)
     *   compute the maximum fusions of solutions in S1 that contain all the
     *   indices in i and prune inferior solutions
     *)
    fun reduceSolns (S1, i) =
      pruneSolns (map (fn s => Reduced (Solution.reduceSoln (getSoln s,i),i,s))
		      S1)

    local
      (* extendSoln' (s1, i)
       *   extend solution s1 under all fusions that are prefixes of getFusion
       *   s1 and construct the set of extended solutions for use as subtree
       *   of a node with indices i
       *)
      fun extendSoln' (s1, i, u) =
	let
	  val s as Solution.S (f, c, frags, rc, t) = getSoln s1
	  val fi = IndexSetSeq.set f
	in
	  if f = IndexSetSeq.any then
	    [Extended (Solution.S (IndexSetSeq.extend (f, i), c, frags, rc, t),
		       IndexSetSeq.extend (f, i), s1)]
	  else if (u orelse length f = 1) andalso
		  (IndexSet.eq (fi, i) orelse IndexSet.subset (i, fi)) then
	    [s1]
	  else
	    map (fn f' => Extended (Solution.extendSoln (s, f', i),
				    IndexSetSeq.extend (f', i),
				    s1))
		(IndexSetSeq.prefixes f)
	end
    in
      (* extendSolns (S1, i, u)
       *   extend the solutions in S1 for tree t to use the indices in
       *   index set i; u is true if the parent node is a unary node;
       *   only extend solutions if i has more indices than t or if the
       *   parent node is not a unary node and there are fusion constraints
       *)
      fun extendSolns (S1, i, u) =
	pruneSolns (SolTreeSet.Union (map (fn s => extendSoln' (s, i, u)) S1))
    end

    (* makeLeafSolns (t, ri)
     *   creates a set of solutions for a leaf node t and
     *   possible redundant indices ri
     *)
    fun makeLeafSolns (t as ExpTree.Array (a, i), ri) =
	  map Leaf (Solution.makeLeafSoln (t, ri))
      | makeLeafSolns (t as ExpTree.Fun (f, i), ri) =
	  map Leaf (Solution.makeLeafSoln (t, ri))
      | makeLeafSolns (t as ExpTree.IntLit i, ri) =
	  map Leaf (Solution.makeLeafSoln (t, ri))
      | makeLeafSolns _ = raise Solution.IllegalExpressionTree

    local
      fun makeUnarySolns' (S1, t, ri) =
	map (fn s => Unary (Solution.makeUnarySoln (getSoln s,t,ri), s)) S1
    in
      (* makeUnarySolns (S1, t, ri)
       *   make unary solutions for an output node or summation node t
       *   out of the extended solutions in set S1 for the subtree of t,
       *   ri contains the possible redundant indices
       *)
      fun makeUnarySolns (S1, t as ExpTree.Output _, ri) =
	    pruneSolns (makeUnarySolns' (S1, t, ri))
	| makeUnarySolns (S1, t as ExpTree.Cast _, ri) =
	    pruneSolns (makeUnarySolns' (S1, t, ri))
	| makeUnarySolns (S1, t as ExpTree.Sum _, ri) =
	    reduceSolns (makeUnarySolns' (S1, t, ri),
			 IndexSet.union (ExpTree.fusible t, ri))
	| makeUnarySolns _ = raise Solution.IllegalExpressionTree
    end

    local
      fun makeBinarySolns' (P, t, ri) =
	map (fn (s1, s2) =>
	       SOME (Binary (Solution.makeBinarySoln (getSoln s1, getSoln s2,
						      t, ri),
			     s1, s2))
		 handle IndexSetSeq.Incompatible => NONE)
	    P

      fun filterSolns nil = nil
	| filterSolns (NONE::t) = filterSolns t
	| filterSolns (SOME s::t) = s :: filterSolns t
    in
      (* makeBinarySolns (S1, S2, t, ri)
       *   creates a set of binary solutions for BinOp node t out of the
       *   solution sets S1 and S2 for subtrees of t and possible redundant
       *   indices ri,
       *   constructs binary solutions for the cross product of S1 and S2,
       *   removes solutions with incompatible fusions and prunes the result,
       *   corresponds to te original MergeSolns()
       *)
      fun makeBinarySolns (S1, S2, t as ExpTree.BinOp (_, _, _), ri) =
	    let val P = SolTreeSet.crossProduct (S1, S2)
	    in
	      pruneSolns (filterSolns (makeBinarySolns' (P, t, ri)))
	    end
	| makeBinarySolns _ = raise Solution.IllegalExpressionTree
    end

    (* minMemFusionSet (t, ri)
     *   runs the fusion algorithm on expTree t and returns a
     *   set of solution trees; the parameter ri contains the set
     *   of possible reduncent indices for space-time tradeoffs,
     *   (ri should really be stored in the tree, but it's easier
     *   to add it here than modify the tree data structure).
     *)
    fun minMemFusionSet (t as ExpTree.Array (a, i), ri) =
	  makeLeafSolns (t, ri)
      | minMemFusionSet (t as ExpTree.Fun (f, i), ri) =
	  makeLeafSolns (t, ri)
      | minMemFusionSet (t as ExpTree.IntLit i, ri) =
	  makeLeafSolns (t, ri)
      | minMemFusionSet (t as ExpTree.Output l, ri) =
	  let
	    val is = ExpTree.indices t
	    val e = extendSolns (minMemFusionSet (l, ri), is, true)
	  in
	    makeUnarySolns (e, t, ri)
	  end
      | minMemFusionSet (t as ExpTree.Cast (_, l), ri) =
	  let
	    val is = ExpTree.indices t
	    val e = extendSolns (minMemFusionSet (l, ri), is, true)
	  in
	    makeUnarySolns (e, t, ri)
	  end
      | minMemFusionSet (t as ExpTree.Sum (_, l), ri) =
	  let
	    val is = ExpTree.indices t
	    val e = extendSolns (minMemFusionSet (l, ri), is, true)
	  in
	    makeUnarySolns (e, t, ri)
	  end
      | minMemFusionSet (t as ExpTree.BinOp (_, l, r), ri) =
	  let
	    val is = ExpTree.indices t
	    val lis = ExpTree.indices l
	    val ris = ExpTree.indices r
	    val lri = IndexSet.union (ri, IndexSet.difference (is, lis))
	    val rri = IndexSet.union (ri, IndexSet.difference (is, ris))
	    val e1 = extendSolns (minMemFusionSet (l, lri), is, false)
	    val e2 = extendSolns (minMemFusionSet (r, rri), is, false)
	  in
	    makeBinarySolns (e1, e2, t, ri)
	  end

    (* exception NoSolution
     *   This exception is thrown if minMemFusion does not find a solution.
     *)
    exception NoSolution

    local
      fun minRecomputation (s, nil) = s
	| minRecomputation (s, h::t) =
	    let
	      val Solution.S (_, _, _, src, _) = getSoln s
	      val Solution.S (_, _, _, hrc, _) = getSoln h
	    in
	      if IntInf.< (src, hrc) then minRecomputation (s, t)
	      else minRecomputation (h, t)
	    end

    in
      (* minMemFusion t
       *   runs the fusion algorithm on expTree t and returns a solution tree
       *)
      fun minMemFusion t =
	let val S = reduceSolns (minMemFusionSet (t, nil), nil)
	in
	  (* After reducing with the empty fusion, all solutions will
	   * have the same fusion.  We will get a set of solutions with
	   * different space-time tradeoffs that all fit within available
	   * memory.  We'll get the one with the least recomputations.
	   * If all solutions for some subtree exceed the memory limit,
	   * the set of solutions will be empty.  In this case, the exception
	   * NoSolution is thrown.
	   *)
	  if length S = 0 then raise NoSolution
	  else minRecomputation (hd S, tl S)
	end
    end

    local
      fun gt (x, y) =
	let
	  val Solution.S (_, (xc, _, _, _), _, _, _) = getSoln x
	  val Solution.S (_, (yc, _, _, _), _, _, _) = getSoln y
	in
	  IntInf.> (xc, yc)
	end

      (* sort s
       *   uses Quicksort to sort the solution tree list s
       *)
      fun sort nil = nil
        | sort (h::t) =
	      sort (List.filter (fn x => gt (x, h)) t)
	        @ [h] @ sort (List.filter (fn x => gt (h, x)) t)
    in
      (* spaceTimeTradeoffs t
       *   runs the fusion algorithm on expTree t and returns all the
       *   space-time tradeoff solutions in decreasing space and
       *   increasing time order
       *)
      fun spaceTimeTradeoffs t =
	sort (reduceSolns (minMemFusionSet (t, nil), nil))
    end
  end
