(***************************************************************************
 *
 * Solution Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   18 September 2001
 * Date:   5 April 2010
 *
 ***************************************************************************)

structure Solution : SOLUTION =
  struct

    type iSet = IndexSetSeq.iSet
    type iSetSeq = IndexSetSeq.iSetSeq
    type expTree = ExpTree.expTree

    (* datatype sol = S of iSetSeq *
     *			   (IntInf.int * IntInf.int * IntInf.int * IntInf.int)
     *			   * (iSet * IntInf.int * IntInf.int * expTree) list *
     *			   IntInf.int * expTree
     * Ex:  S ([["i", "j"]], (1,1,0,0), [], 0,
     *		Array ("A", [("i",500), ("j",100)]))
     *
     * S (f, (hi, lo, hiC, loC), frags, rc, t)
     *   Constructs a solution with fusion f, high memory cost hi,
     *   low memory lost lo, high memory cost of children hiC,
     *   low memory cost of children loC, list of fragments frags,
     *   recomputation cost rc, and expression tree t.
     *
     *   For the static memory allocation cost model, the memory cost is in hi,
     *   lo = fusedSize(f, t), hiC = loC = 0.
     *
     *   For the dynamic memory allocation cost model,
     *   hiC is the largest amount of memory needed to evaluate the subtrees,
     *   loC is the memory cost for the sum of the root nodes of the subtrees,
     *   for the subtrees that are fused into the root fragment.
     *
     *   lo(t,f)	the memory needed for the root of t under fusion f
     *   hi(t,f)	the maximal memory needed for tree t under fusion f
     *
     *   frags		a (iSet,hf,lf,p) list representing code fragments
     *			that haven't been fused with the root
     *			fragments are sorted in non-increasing iSet size;
     *			each fragment contains
     *     iSet		the set of fused indices for the fragment
     *     hf		the hi of the fragment
     *     lf		the lo of the fragment
     *     p		the parent of the fragment
     *
     *   lo(t,f) = fusedSize(t,f)
     *   hi(t,f) = max(hiC(t), loC(t) + lo(t,f))
     *
     *   loC(t)  = sum_i (lo(c_i, fusion(c_i,t))) for all children c_i of t
     *		   that are fused with t
     *   hiC(t)  = max_i (hi(c_i,fusion(c_i,t)) + sum_j(lo(c_j,fusion(c_j,t))))
     *		   for all children c_i of t that are fused with t and j < i,
     *		   where the children are sorted in decreasing
     *		   hi(c_i,fusion(c_i,t)) - lo(c_i,fusion(c_i),t) order
     *
     *   where	   fusion(c,t) is the actual fusion between child c of t and t
     *)
    datatype sol = S of iSetSeq *
			(IntInf.int * IntInf.int * IntInf.int * IntInf.int) *
			(iSet * IntInf.int * IntInf.int * expTree) list *
			IntInf.int * expTree

    (* getFusion s
     *   gets the fusion out of solution s
     *)
    fun getFusion (S (f, _, _, _, _)) = f

    (* exception IllegalExpressionTree
     *   this exception is thrown if a function is applied to an
     *   expression tree it does not apply to; this exception is used
     *   as a run-time sanity check: if the program is correct, it
     *   should never be thrown
     *)
    exception IllegalExpressionTree

    (* DetailedCostModel
     *   This variable is used to select the cost model.  If true,
     *   the cost model does not include the cost of scalar temporaries.
     *   If false, the cost model assumes that every temporary, even if
     *   is is a scalar, will be stored in memory.
     *
     *   When creating the solution for a non-leaf node, the cost of this
     *   node is calculated as 1 since, potentially, it could be fully fused
     *   with its parent.  The cost for the subtrees must be adjusted.
     *   In the simple cost model, the cost of 1 that was added previously
     *   is replaced with the fusedSize of the node when extending the
     *   solution of the subtree to the fusion of the parent (in extendSoln).
     *   In the detailed cost model, the same thing happens, but if the
     *   fusedSize is still 1, a cost of 0 is budgeted since the temporary
     *   holding the result of the subtree is a scalar and can be eliminated.
     *
     *   The only functions effected by this variable are makeUnarySoln
     *   and makeBinarySoln.
     *)
    val DetailedCostModel = ref true

    (* setCostModelDetails n
     *   Queries and selects the cost model details.  If n is NONE, it
     *   returns the value of DetailedCostModel, if n is (SOME b), it sets
     *   DetailedCostModel to b and returns the old value.
     *)
    fun setCostModelDetails NONE = !DetailedCostModel
      | setCostModelDetails (SOME b) =
	  let val old = !DetailedCostModel
	  in
	    DetailedCostModel := b;
	    old
	  end

    (* DynamicMemAllocCostModel
     *   This variable is used to select the cost model.  If true,
     *   the cost model assumes that child nodes are deallocated after
     *   the parent node has been computed.  If false, the cost model
     *   assumes that temporaries will remain in memory, even after
     *   their value has been used.  The selection of a DetailedCostModel
     *   affect both the static and dynamc memory allocation cost models.
     *)
    val DynamicMemAllocCostModel = ref false

    (* setCostModelAllocation n
     *   Queries and selects the cost model.  If n is NONE, it returns
     *   the value of DynamicMemAllocCostModel, if n is (SOME b), it sets
     *   DynamicMemAllocCostModel to b and returns the old value.
     *)
    fun setCostModelAllocation NONE = !DynamicMemAllocCostModel
      | setCostModelAllocation (SOME b) =
	  let val old = !DynamicMemAllocCostModel
	  in
	    DynamicMemAllocCostModel := b;
	    (* FIX: The dynamic memory allocation cost model is not
	     * net implmented with a detailed cost accounting for scalars
	     *)
	    if b then DetailedCostModel := false else ();
	    old
	  end

    (* SpaceTimeTradeoffs
     *   This variable is used to select the cost model.  If it refers
     *   to NONE, we do not perform space-time tradeoffs, i.e., we assume
     *   we have unlimited memory.  If it refers to SOME n, where n is
     *   an infinite-precision integer, we perform space-time tradeoffs
     *   for maximum memory size n.  The default value is ref NONE.
     *)
    val SpaceTimeTradeoffs : IntInf.int option ref = ref NONE

    (* setMemoryLimit n
     *   Turns on space-time tradeoffs and sets the memory limit to n.
     *)
    fun setMemoryLimit n = SpaceTimeTradeoffs := SOME n

    (* clearMemoryLimit ()
     *   Turns off space-time tradeoffs.
     *)
    fun clearMemoryLimit () = SpaceTimeTradeoffs := NONE

    (* exceedsMemoryLimit n
     *   Returns true iff space-time tradeoffs are enabled and n
     *   exceeds the memory limit.
     *)
    fun exceedsMemoryLimit n =
      case !SpaceTimeTradeoffs of
	  NONE => false
	| SOME limit => IntInf.> (n, limit)

    (* FunctionCost
     *   This variable contains the cost of a function call.  When
     *   performing space-time tradeoffs, we try to minimize the
     *   recomputation cost.  The cost of recomputing an array element
     *   in a unary or binary operation is 1.  The cost of recomputing
     *   a function is functionCost.  Ideally, this information should
     *   be declared in the input file on a per-function basis.  For
     *   this prototype we chose to make it constant for all functions.
     *)
    val FunctionCost = ref (IntInf.fromInt 1000)

    (* setFunctionCost n
     *   Sets the function call cost for space-time tradeoffs to n.
     *)
    fun setFunctionCost n = FunctionCost := n

    local
      fun toStringFrag (i,h,l,_) =
	  "(" ^ IndexSet.toString i ^ ","
	  ^ IntInf.toString h ^ ","
	  ^ IntInf.toString l ^ ")"

      fun toString' nil = ""
	| toString' (frag :: nil) =
	  toStringFrag frag
	| toString' (frag :: t) =
	  toStringFrag frag ^ ", " ^ toString' t
    in
      (* toString s
       *   creates a string representation of solution s
       *)
      fun toString (S (f, (hi, lo, hiC, loC), frags, rc, t)) =
          "S (" ^ IndexSetSeq.toString f ^ ", "
	  ^ (if !DynamicMemAllocCostModel then
		 "(" ^ IntInf.toString hi ^ ","
		 ^ IntInf.toString lo ^ ","
		 ^ IntInf.toString hiC ^ ","
		 ^ IntInf.toString loC ^ "), ["
		 ^ toString' frags ^ "], "
	     else
		 IntInf.toString hi ^ ", ")
          ^ (if !SpaceTimeTradeoffs = NONE then ""
	     else IntInf.toString rc ^ ", ")
	  ^ ExpTree.toExp t ^ ")\n"
    end

    structure SolElem =
      struct
	type t = sol
	val eq = op=
	val toString = toString
      end

    structure SolSet = Set (structure Elem = SolElem)

    type solSet = SolSet.t SolSet.set

    (* max (n, m)
     *   Returns the maximum of the two IntInf integers n and m.
     *)
    fun max (n, m) = if IntInf.> (n, m) then n else m

    (* zero
     *   The IntInf constant 0.
     *)
    val zero = IntInf.fromInt 0

    (* one
     *   The IntInf constant 1.
     *)
    val one = IntInf.fromInt 1

    (* mergeFrags (frags, frags')
     *   merge two lists of frags in non-decreasing set size
     *)
    fun mergeFrags (nil, frags') = frags'
      | mergeFrags (frags, nil) = frags
      | mergeFrags (frags as (fi,hi,lo,t) :: tf,
		    frags' as (fi',hi',lo',t') :: tf') =
	  if ((fi <> fi') andalso IndexSet.subset (fi', fi))
	     orelse
	     (fi = fi' andalso
	      IntInf.>= (IntInf.- (hi', lo'), IntInf.- (hi, lo))) then
	    (fi,hi,lo,t) :: mergeFrags (tf, frags')
	  else
	    (fi',hi',lo',t') :: mergeFrags (frags, tf')

    (* combineFrags frags
     *   combine consecutive frags with equal fused set into one frag
     *)
    fun combineFrags nil = nil
      | combineFrags (frags as _::nil) = frags
      | combineFrags ((fi,hi,lo,t) :: (fi',hi',lo',t') :: tail) =
	  if fi = fi' then
	    combineFrags ((fi, max (hi',lo'+hi), lo+lo', t) :: tail)
	  else
	    (fi,hi,lo,t) :: combineFrags ((fi',hi',lo',t') :: tail)

    (* low frags
     *   adds up all the lo's in the fragments
     *)
    fun low nil = 0
      | low ((_,_,lo,_) :: t) = IntInf.+ (lo, low t)

    (* leqFrags (frags, frags')
     *   returns true iff frags is inferior or equal to frags'
     *)
    fun leqFrags (_, nil) = true
      | leqFrags (nil, _) = false
      | leqFrags (frags as (fi,hi,_,_) :: t, frags' as (fi',hi',_,_) :: t') =
	  if fi = fi' andalso (IntInf.< (hi, hi') orelse
			       IntInf.< (low t, low t')) then false
	  else if fi = fi' then leqFrags (t, t')
	  else if IndexSet.subset (fi', fi) then leqFrags (t, frags')
	  else leqFrags (frags, t')

    (* init
     *   a marker tree node to indicate that a fragment is an initialization
     *)
    val init = ExpTree.IntLit zero

    (* useFrags (sol, e)
     *   use all the fragments in solution sol with equal or larger
     *   set of fused indices and fuse them with the root fragment;
     *   if useFrags is called from extendSoln (if e is true), combine
     *   any fragments with equal fused set
     *)
    fun useFrags (s as S (_, _, nil, _, _), _) = s
      | useFrags (S (f, (hi,lo,hiC,loC),
		     frags as (fi,hi',lo',t')::tail, rc, t), e) =
	  if t' = init then	(* hd frags is initialization of t *)
	    useFrags (S (f, (lo'+hi,lo',hiC,loC), tail, rc, t), e)
	  else if fi = (IndexSetSeq.set f) orelse
		  IndexSet.subset (IndexSetSeq.set f, fi) then
	    let
	      val loC' = loC + lo'
	      val hiC' = max (hi', lo' + hiC)
	    in
	      if t' = t andalso	(* hd frags is a child of t *)
		 not (fi = (IndexSetSeq.set f)) then
		let
		  val hi'' = max (hiC', loC' + lo)
		in
		  useFrags (S (f, (hi'',lo,hiC',loC'), tail, rc, t), e)
		end
	      else		(* hd frags is a decendent of t *)
		let
		  val hi'' = max (hi', lo' + hi)
		in
		  useFrags (S (f, (hi'',lo,hiC',loC'), tail, rc, t), e)
		end
	    end
	  else
	    S (f, (hi,lo,hiC,loC),
	       if e then combineFrags frags else frags,
	       rc, t)

    (* makeLeafSoln (t, ri)
     *   creates a solution set for a leaf node t and possible redundant
     *   indices ri;
     *   arrays and integer literals can be read at zero cost and can
     *   have any nesting, function calls have memory cost 1;
     *   there is no difference between static and dynamic allocation;
     *   corresponds to code in InitSolns(), InitMemUsage(), and
     *   MergeMemUsage() from the original algorithm
     *)
    fun makeLeafSoln (t as ExpTree.IntLit _, ri) =
	  (case !SpaceTimeTradeoffs of
	       NONE =>
		 [S (IndexSetSeq.any, (zero, zero, zero, zero), [], zero, t)]
	     | SOME limit =>
		 let
		   val fs = map (fn ri' => if ri' = nil then [] else [ri'])
				(IndexSet.powerSet ri)
		 in
		   map (fn f => S (f, (zero, zero, zero, zero), [], zero, t))
		       fs
		 end
	  )
      | makeLeafSoln (t as ExpTree.Array _, ri) =
	  (case !SpaceTimeTradeoffs of
	       NONE =>
		 [S (IndexSetSeq.any, (zero, zero, zero, zero), [], zero, t)]
	     | SOME limit =>
		 let
		   val i = ExpTree.fusible t
		   val fs = map (fn ri' =>
				   let val i' = IndexSet.union (i, ri')
				   in
				     if i' = nil then [] else [i']
				   end)
				(IndexSet.powerSet ri)
		 in
		   map (fn f => S (f, (zero, zero, zero, zero), [], zero, t))
		       fs
		 end
	  )
      | makeLeafSoln (t as ExpTree.Fun (_, _), ri) =
	  (case !SpaceTimeTradeoffs of
	       NONE =>
		 let
		   val f = [ExpTree.fusible t]
		   val c = ExpTree.fusedSize (t, f)
		 in
		   [S (f, (c, c, zero, zero), [], zero, t)]
		 end
	     | SOME limit =>
		 let
		   val i = ExpTree.fusible t
		   val fs = map (fn ri' =>
				   let val i' = IndexSet.union (i, ri')
				   in
				     if i' = nil then [] else [i']
				   end)
				(IndexSet.powerSet ri)
		 in
		   map (fn f =>
			   let
			     val c = ExpTree.fusedSize (t, f)
			   in
			     S (f,
				(c, c, zero, zero), [],
				 IntInf.* (ExpTree.recomputationCost (t, f),
					   !FunctionCost),
				t)
			   end)
		       fs
		 end
	  )
      | makeLeafSoln _ = raise IllegalExpressionTree

    (* makeUnarySoln (s, t, ri)
     *   constructs the solution for an output node or summation node t
     *   out of a subtree solution s and possible redundant indices ri
     *)
    fun makeUnarySoln (S (f, (hiC, loC, _, _), frags, rc, _),
		       t as ExpTree.Output _, ri) =
	  if !DynamicMemAllocCostModel then
	    S (f, (hiC, zero, hiC, loC), frags, rc, t)	    
	  else
	    S (f, (hiC, zero, zero, zero), frags, rc, t)
      | makeUnarySoln (S (f, (hiC, loC, _, _), frags, rcC, l),
		       t as ExpTree.Cast (_, _), ri) =
	  let
	    val i = ExpTree.indices t
	    val ef = IndexSetSeq.extend (f, i)
	    val sf = IndexSetSeq.set f
	    val lo = ExpTree.fusedSize (t, [i])
	  in
	    if !DynamicMemAllocCostModel then
	      S (ef, (lo, lo, zero, zero), (sf, hiC, loC, t) :: frags, rcC, t)
	    else
	      let
		val hi = IntInf.+ (hiC, lo)
	      in
	        S (ef, (hi, lo, zero, zero), [], rcC, t)
	      end
	  end
      | makeUnarySoln (S (f, (hiC, loC, _, _), frags, rcC, l),
		       t as ExpTree.Sum _, ri) =
	  let
	    val i = ExpTree.indices t
	    val ef = IndexSetSeq.extend (f, i)
	    val sf = IndexSetSeq.set f
	    val rc = IntInf.+ (rcC, ExpTree.recomputationCost (t, f))
	    val dc = !DetailedCostModel andalso loC = one andalso sf = i
	    val lo = ExpTree.fusedSize (t, [i])
	  in
	    if !DynamicMemAllocCostModel then
	      if sf = i then
	        let
		  val hi = max (hiC, if dc then lo else IntInf.+ (loC, lo))
	        in
		  S (ef, (hi, lo, hiC, if dc then zero else loC), frags, rc, t)
	        end
	      else
		S (ef, (lo, lo, zero, zero), (sf, hiC, loC, t) :: frags, rc, t)
	    else
	      let
		val hi = if dc then hiC else IntInf.+ (hiC, one)
	      in
		S (ef, (hi, lo, zero, zero), [], rc, t)
	      end
	  end
      | makeUnarySoln _ = raise IllegalExpressionTree

    (* makeBinarySoln (sL, sR, t, ri)
     *   merges solutions sL and sR of children of t and adds possible
     *   redundant indices ri, throws exception IndexSetSeq.Incompatible
     *   if the fusions in sL and sR can't be merged;
     *   needs extended solutions esL and esR for calculating nesting
     *   and non-extended solutions for calculating cost;
     *   corresponds to code in MergeSolns() and MergeMemUsage()
     *)
    fun makeBinarySoln (S (fL, (hiL, loL, _, _), fragsL, rcL, l),
			S (fR, (hiR, loR, _, _), fragsR, rcR, r),
			t as ExpTree.BinOp _, ri) =
	  let
	    (* Recompute ef and ef' instead of passing the extended
	     * solutions es and es' from SolTree.makeBinarySoln'. *)
	    val i = ExpTree.indices t
	    val efL = IndexSetSeq.extend (fL, i)
	    val efR = IndexSetSeq.extend (fR, i)
	    val sfL = IndexSetSeq.set fL
	    val sfR = IndexSetSeq.set fR

	    val cl = if !DetailedCostModel andalso IntInf.> (hiL, zero)
			  andalso loL = one andalso sfL = i
			  andalso ExpTree.indices l = ExpTree.fusible l
		       then one else zero
	    val cr = if !DetailedCostModel andalso IntInf.> (hiR, zero)
			  andalso loR = one andalso sfR = i
			  andalso ExpTree.indices r = ExpTree.fusible r
		       then one else zero
	    val f = IndexSetSeq.merge (efL, efR)
	    val rc = IntInf.+ (ExpTree.recomputationCost (t, f),
			       IntInf.+ (rcL, rcR))
	    val lo =  ExpTree.fusedSize (t, f)
	  in
	    if !DynamicMemAllocCostModel then
	      if sfL = i andalso sfR = i then
		(* both subtrees are fused with parent *)
		let
		  val hiC = if IntInf.>= (IntInf.- (hiL, loL),
					  IntInf.- (hiR, loR)) then
			      max (hiL, IntInf.+ (loL, hiR))	(* eval L, R *)
			    else
			      max (hiR, IntInf.+ (loR, hiL))	(* eval R, L *)
		  val loC = IntInf.+ (IntInf.- (loL, cl), IntInf.- (loR, cr))
		  val hi = max (hiC, IntInf.+ (loC, lo))
		in
		  S (f, (hi,lo,hiC,loC), mergeFrags (fragsL, fragsR), rc, t)
		end
	      else if sfL = i then
		(* only left subtree is fused *)
		let
		  val hi = max (hiL, IntInf.+ (IntInf.- (loL, cl), lo))
		  val fragsR' = (sfR, hiR, loR, t) :: fragsR
		in
		  S (f, (hi,lo,hiL,loL), mergeFrags (fragsL, fragsR'), rc, t)
		end
	      else if sfR = i then
		(* only right subtree is fused *)
		let
		  val hi = max (hiR, IntInf.+ (IntInf.- (loR, cr), lo))
		  val fragsL' = (sfL, hiL, loL, t) :: fragsL
		in
		  S (f, (hi,lo,hiR,loR), mergeFrags (fragsL', fragsR), rc, t)
		end
	      else
		let
		  val fragsL' = (sfL, hiL, loL, t) :: fragsL
		  val fragsR' = (sfR, hiR, loR, t) :: fragsR
		in
		  S (f, (lo,lo,zero,zero), mergeFrags (fragsL',fragsR'), rc, t)
		end
	    else
	      let
		val hi = IntInf.+ (IntInf.+ (IntInf.- (hiL, cl),
					     IntInf.- (hiR, cr)), one)
	      in
	        S (f, (hi, lo, zero, zero), [], rc, t)
	      end
	  end
      | makeBinarySoln _ = raise IllegalExpressionTree

    (* extendSoln (s, f', i)
     *   extend solution s under fusion f' (which is a prefix of getFusion s)
     *   and construct a solution for use as subtree of a node with indices i
     *)
    fun extendSoln (S (f, (hi, lo, hiC, loC), frags, rc, t), f', i) =
      let
	val lo' = ExpTree.fusedSize (t, f')
      in
	if !DynamicMemAllocCostModel then
	  let
	    val fi = IndexSetSeq.set f'
	    val ti = ExpTree.indices t
	  in
	    if fi = ti then
	      S (f', (hi,lo,hiC,loC), frags, rc, t)
	    else
	      let
		val frags' = mergeFrags ([(fi,lo',lo',init)], frags)
	      in
	        useFrags (S (f', (hiC,zero,hiC,loC), frags', rc, t), true)
	      end
	  end
	else
	  let
	    val hi' = IntInf.+ (IntInf.- (hi, lo), lo')
	  in
	    S (f', (hi', lo', zero, zero), [], rc, t)
	  end
      end

    (* reduceSoln (s, i)
     *   reduces the solution s to the maximum fusion with index set i
     *)
    fun reduceSoln (S (f, (hi, lo, hiC, loC), frags, rc, t), i) =
      let
	val f' = IndexSetSeq.maxPrefix (f, i)
	val lo' = ExpTree.fusedSize (t, f')
      in
	if !DynamicMemAllocCostModel then
	  let
	    val fi = IndexSetSeq.set f'
	    val frags' = mergeFrags ([(fi,lo',lo',init)], frags)
	  in
	    useFrags (S (f', (hiC,zero,hiC,loC), frags', rc, t), false)
	  end
	else
	  let
	    val hi' = IntInf.+ (IntInf.- (hi, lo), lo')
	  in
	    S (f', (hi', lo', zero, zero), [], rc, t)
	  end
      end

    (* inferior (s, s')
     *   returns true if solution s is inferior to solution s'
     *)
    fun inferior (S (fL, (hiL, loL, _, _), fragsL, rcL, _),
		  S (fR, (hiR, loR, _, _), fragsR, rcR, _)) =
      IntInf.>= (hiL, hiR) andalso IntInf.>= (rcL, rcR)
        andalso IndexSetSeq.leq (fL, fR)
        andalso (not (!DynamicMemAllocCostModel) orelse
		 (IntInf.>= (loL, loR)
		  andalso IntInf.>= (low fragsL, low fragsR)
		  andalso leqFrags (fragsL, fragsR)))

  end
